use microrm::{schema::entity::Entity, prelude::{Queryable, Insertable}};

mod clap_interface;

pub use clap_interface::ClapInterface;

mod role;
mod user;

pub trait Object: Sized + Entity + std::fmt::Debug {
    type Error;
    type CreateParameters: clap::Parser + std::fmt::Debug;
    fn create_from_params(_: &Self::CreateParameters) -> Result<Self, Self::Error>;

    type ExtraCommands: clap::Subcommand + std::fmt::Debug;
    type ExtraCommandData;
    fn run_extra_command(
        data: &Self::ExtraCommandData,
        extra: &Self::ExtraCommands,
        query_ctx: impl Queryable<EntityOutput = Self>,
        insert_ctx: &impl Insertable<Self>) -> Result<(), Self::Error> { unreachable!() }

    fn shortname(&self) -> &str;
}

#[derive(Debug)]
pub struct EmptyCommand;

impl clap::FromArgMatches for EmptyCommand {
    fn from_arg_matches(matches: &clap::ArgMatches) -> Result<Self, clap::Error> {
        Ok(EmptyCommand)
    }

    fn update_from_arg_matches(&mut self, matches: &clap::ArgMatches) -> Result<(), clap::Error> {
        Ok(())
    }
}

impl clap::Subcommand for EmptyCommand {
    fn augment_subcommands(cmd: clap::Command) -> clap::Command {
        cmd
    }

    fn augment_subcommands_for_update(cmd: clap::Command) -> clap::Command {
        cmd
    }

    fn has_subcommand(name: &str) -> bool { false }
}
